import os

from llama_index.core.llms.llm import LLM
from llama_index.core.tools.tool_spec.base import BaseToolSpec

from qai.agent import ROOT_DIR
from qai.agent.models import OutreachType

PROJ_DIR = os.path.dirname(os.path.dirname(os.path.dirname(ROOT_DIR)))


class OutreachToolSpec(BaseToolSpec):
    """Simple Find Outreach tool."""

    spec_functions = ["find_outreach"]

    def __init__(self, llm: LLM = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.llm = llm

    def find_outreach(self, outreach_types: list[OutreachType]) -> list[OutreachType]:
        """
        Find the correct outreach types given a list of outreach types.
        Args:
            outreach_types: The list of outreach types to find the correct outreach type for.
        Returns:
            The list of outreach types.
        """
        ## Input is not perfect from the llm, it can be a list of steps but in dict form
        ## or it can be a dict with a list of dicts
        ## comes in as a list of steps but in dict form
        if isinstance(outreach_types, dict):
            outreach_types = outreach_types["outreach_types"]
        outreach_types = [OutreachType(**outreach_type) for outreach_type in outreach_types]
        outreach_types = self._refine_step_queries(outreach_types)
        print(f"Returning types = {outreach_types}")
        return outreach_types



from pprint import pformat

from flask import jsonify, request
from flask_pydantic import validate
from pi_log import logs
from qai.chat import VERSION, create_app
from qai.chat.config import cfg
from qai.chat.models import CompanyQueryModel

app = create_app(cfg)  ## app should be created before importing other modules


from qai.chat.chat import company_query
from qai.chat.db.chroma.chroma import Chroma
from qai.chat.server.security import requires, token_required

log = logs.getLogger(__name__)
logs.set_app_level("debug")
default_not_found_message = "Sorry, I don't know the answer to that question."


@app.route("/heartbeat", methods=["GET", "POST"])
@requires(["key", "value"])
def heartbeat():
    log.debug(f"serve::heartbeat")
    return jsonify({"status": "OK"}), 200


@app.route("/set_config", methods=["GET", "POST"])
@token_required
@requires(["key", "value"])
def set_config():
    in_data = request.get_json()
    key = in_data.get("key")
    value = in_data.get("value")
    log.debug(f"serve::set_config, in_data={in_data}")
    old = app.cfg[key]
    app.cfg[key] = value
    return jsonify({"old": old, "new": value}), 200


@app.route("/set_loglevel/<level>", methods=["GET", "POST"])
@token_required
def set_loglevel(level: str):
    logs.set_log_level(level)
    return jsonify({"status": "OK", "message": f"log level set to {level}"}), 200


@app.route("/version", methods=["GET", "POST"])
@token_required
def version():
    log.debug(f"serve::version")
    return jsonify({"config_version": app.cfg.version, "app_version": VERSION}), 200


@app.route("/get_collection", methods=["GET", "POST"])
@token_required
def chroma():
    in_data = request.get_json()
    log.debug(f"serve::chroma, in_data={in_data}, chroma.cfg={app.cfg.chroma}")
    c = Chroma(app.cfg.chroma)
    col = c.get_collection(in_data["collection_name"])
    count = col.count() if col else -1
    meta = col.metadata if col else {}
    log.debug(f"serve::chroma count={count}, metadata={meta}")
    return jsonify({"count": count, "in_data": in_data, "metadata": meta}), 200


@app.route("/config", methods=["GET", "POST"])
@token_required
def config():
    log.debug(f"serve::config \n{pformat(app.cfg)}")
    return (
        jsonify(
            {
                "config": app.cfg,
            }
        ),
        200,
    )


def _debug_query(in_data, guid: str, query: str, company_name: str):
    query = query[36:].strip().lstrip(":").strip()

    in_data.pop("query")
    d = {
        "app_version": VERSION,
        "config_version": app.cfg.version,
        "build_time": app.cfg.build_time,
        "query": query,
        "in_data": in_data,
    }
    in_data["response"] = pformat(d)
    return jsonify(in_data), 200


@app.route("/", methods=["GET", "POST"])
@token_required
@validate(body=CompanyQueryModel)
def company_chat():
    model = CompanyQueryModel(**request.get_json())
    model._fill_company_info(app.website_2_id, app.id_2_website)

    log.debug(f"serve::query, in_data={model}")

    if model.mock:
        return {"response": "Mocked Response"}, 200
    response = company_query(
        query=model.query,
        user_id=model.user_id,
        company_name=model.company_name,
        company_id=model.company_id,
    )
    response = {"response": str(response)}
    return jsonify(response), 200


if __name__ == "__main__":
    app.run(host="0.0.0.0", debug=True)



from typing import Self

from llama_index.agent.openai import OpenAIAgent
from llama_index.core.memory import ChatMemoryBuffer
from llama_index.core.memory.chat_memory_buffer import ChatMemoryBuffer
from llama_index.core.storage.chat_store import SimpleChatStore
from llama_index.llms.openai import OpenAI
from pi_conf import load_config
from pi_log import logs

from qai.agent.agents.find_agent import AgentType, FindAgentWorker
from qai.agent.sessions.qai_session import QaiSession

log = logs.getLogger(__name__, to_stdout=True, level=logs.DEBUG)
TOKEN_LIMIT = 3000


class QaiBot(OpenAIAgent):

    def __init__(
        self,
        # db_config: dict,
        # model_config: dict,
        # tools: List[BaseTool],
        llm: OpenAI = None,
        model_config: dict = None,
        db_config: dict = None,
        *args,
        **kwargs,
    ):
        self.llm = llm
        self.chat_store = SimpleChatStore()
        self.model_config = model_config
        self.db_config = db_config
        self.last_sessions: dict[str, QaiSession] = {}
        self.sessions: dict[str, QaiSession] = {}

    @classmethod
    def create(cls, db_config: dict, model_config: dict, *args, **kwargs) -> Self:
        llm = OpenAI(**model_config)
        qaibot = QaiBot(
            llm=llm,
            db_config=db_config,
            model_config=model_config,
            *args,
            **kwargs,
        )
        return qaibot

    def chat(
        self, query: str, user_id: str, session_id: str = None, token_limit: int = TOKEN_LIMIT
    ):
        chat_memory = ChatMemoryBuffer.from_defaults(
            token_limit=token_limit,
            chat_store=self.chat_store,
            chat_store_key=user_id,
        )
        from pprint import pprint

        pprint(chat_memory.get_all())

        agent = FindAgentWorker.create(chat_memory=chat_memory)
        agent_type = agent.chat(query)
        print(f"Agent type: {agent_type}")
        session = self._get_session(agent_type, session_id, user_id=user_id)
        if session is None:
            session = QaiSession(
                db_config=self.db_config,
                model_config=self.model_config,
                session_id=session_id,
                agent_type=agent_type,
                memory=chat_memory,
            )
            self.sessions[session.session_id] = session
        self.last_sessions[user_id] = session
        return session.chat(query)

    def _get_session(self, agent_info: AgentType, session_id: str, user_id: str) -> QaiSession:
        if session_id in self.sessions:
            return self.sessions[session_id]
        if self.last_sessions.get(user_id) is None:
            return None
        agent_type = agent_info
        last_session = self.last_sessions.get(user_id)
        if agent_type == last_session.agent_type:
            last_session
        return None


if __name__ == "__main__":
    cfg = load_config("qai-chat")
    cfg.to_env()
    agent = QaiBot.create(db_config=cfg.get("db_config"), model_config=cfg.model)
    sentence = "I want to create a campaign"
    print(agent.chat(sentence, user_id="1"))
    sentence = "With lots of toppings"
    print(agent.chat(sentence, user_id="1"))

    sentence = "and more"
    print(agent.chat(sentence, user_id="1"))

    print("FINISHED")



from typing import Optional, TypeVar

from llama_index.agent.openai import OpenAIAgent
from llama_index.core.agent.runner.base import AgentRunner
from llama_index.core.base.llms.types import ChatMessage, MessageRole
from llama_index.core.llms import ChatMessage
from llama_index.core.llms.llm import LLM
from llama_index.core.memory import ChatMemoryBuffer
from llama_index.core.memory.chat_memory_buffer import ChatMemoryBuffer
from llama_index.core.memory.types import BaseMemory
from llama_index.core.storage.chat_store import SimpleChatStore
from llama_index.core.tools.tool_spec.base import BaseToolSpec
from llama_index.llms.openai import OpenAI
from pi_conf import load_config

from qai.ai import MessageRole
from qai.ai.frameworks.openai.llm import OpenAILLM

T = TypeVar("T", AgentRunner)


class ConversationToolSpec(BaseToolSpec):
    name: str = "conversational_chatbot"
    description: str = "A conversational chatbot agent."
    spec_functions: list[str] = ["converse"]
    memory: BaseMemory = None
    user_id: str = None

    def __init__(
        self, user_id: str, model_config: dict, memory: BaseMemory = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.user_id = user_id
        self.memory = memory
        self.model_config = model_config

    def converse(self, query: str) -> str:
        print(f"Conversing... user_id: {self.user_id}, query: {query}")
        self.memory.chat_store.add_message(
            key=self.user_id,
            message=ChatMessage(
                content=query,
                role=MessageRole.USER,
            ),
        )
        msgs = self.memory.chat_store.get_messages(self.user_id)
        msgs = [{"role": msg.role, "content": msg.content} for msg in msgs]
        print("Conversing...", msgs)
        llm = OpenAILLM(config=self.model_config)
        qr = llm.query(query, messages=msgs)

        self.memory.chat_store.add_message(
            key=self.user_id,
            message=ChatMessage(**qr.to_message()),
        )

    def to_agent(
        self,
        llm: Optional[LLM],
        verbose: bool = False,
        agent_cls: type[T] = OpenAIAgent,
    ) -> T:
        agent = agent_cls.from_tools(
            tools=self.to_tool_list(),
            llm=llm,
            verbose=verbose,
        )
        return agent


if __name__ == "__main__":
    cfg = load_config("qai-chat")
    cfg.to_env()
    llm = OpenAI(**cfg.model)
    chat_store = SimpleChatStore()
    user_id = "1234"
    chat_memory = ChatMemoryBuffer.from_defaults(
        token_limit=500,
        chat_store=chat_store,
        chat_store_key=user_id,
    )

    conversation_spec = ConversationToolSpec(
        user_id=user_id,
        memory=chat_memory,
        model_config=cfg.model,
    )
    agent = conversation_spec.to_agent(llm=llm, verbose=True)
    agent.chat("Hello, how are you?")
    agent.chat("Can you tell me more?")
    from pprint import pprint

    print("@@@@@@@@@@@@@@@@@")
    pprint(agent.memory.chat_store.get_messages(user_id))
    # print(conversation_tool.chat("Hello, how are you?"))
    # print(conversation_tool.create(nametools=None, tools=None))



import os
from typing import List, Optional

from llama_index.core import SQLDatabase
from llama_index.core.base.response.schema import Response
from llama_index.core.bridge.pydantic import BaseModel, Field
from llama_index.core.llms.llm import LLM
from llama_index.core.query_engine import NLSQLTableQueryEngine
from llama_index.core.tools.types import ToolOutput
from llama_index.tools.database import DatabaseToolSpec
from qai.ai import QueryReturn
from sqlalchemy import create_engine, text
from sqlalchemy.exc import NoSuchTableError
from sqlalchemy.schema import CreateTable

from qai.agent import ROOT_DIR

# Current supported fields in backend: email, name, phone_number, company_name, company_url, job_title, linkedin_url, timezone

PROJ_DIR = os.path.dirname(os.path.dirname(os.path.dirname(ROOT_DIR)))


# Settings.llm = OpenAI(model="gpt-4", temperature=0.0)


class PeopleIds(BaseModel):
    id: str = Field(..., description="The id of the person")


class CompanyIds(BaseModel):
    id: str = Field(..., description="The id of the company")


class IndustryIds(BaseModel):
    id: str = Field(..., description="The id of the industry")


class StepModel(BaseModel):
    sentence: str = Field(
        ...,
        description=(
            "A revised sentence that captures the business logic. the revised sentence keeps details such as numbers."
        ),
    )
    category: str = Field(
        ..., description="The category of the step, for example: company, people, etc."
    )

    class Config:
        arbitrary_types_allowed = True


class RefineSQLQuery(DatabaseToolSpec):
    """Simple Database tool."""

    ## List of tables in the database, None for all tables
    tables: Optional[list[str]] = None
    llm: Optional[LLM] = None
    limit: int = None  ## Limit the number of results returned

    refine_query_system_message: str = (
        "Separate the text into a list of logical execution steps for a business user and categorize the step."
        "Do not create steps, only refine the given sentences into logical execution steps."
        "The steps should be in the order of execution. "
        "If there are no steps ask for clarification."
        "Keep details such as quantities or numbers."
        "When choosing a category only use one of the following: [company, people, industry, location, time, campaign, email, extraneous]"
        "For example if the user query is 'Get me the head of sales for the top 5 companies in the seattle area' and the following database schema {schema}, the refined query should be: "
        "step: Get the top 5 companies in the seattle area, category: company"
        "step: Get the head of sales for each company, category: people"
        ""
    )
    spec_functions = [
        "find_steps",
        "load_data",
        "describe_tables",
        "list_tables",
        "load_people",
        "load_companies",
    ]

    def __init__(self, llm: LLM = None, limit: int = 5, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.llm = llm
        self.limit = limit
        ## get columns from the database
        self.columns = self.list_tables
        # self.spec_functions.extend(self.extra_spec_functions)

    def tables_to_json(self, tables: Optional[List[str]] = None) -> dict:
        """
        Converts the specified tables in the database into a simple json format

        Args:
            tables (List[str]): A list of table names to retrieve details about
        """
        table_names = tables or [table.name for table in self._metadata.sorted_tables]
        json_tables = {}
        for table_name in table_names:
            columns = []
            table = next(
                (table for table in self._metadata.sorted_tables if table.name == table_name),
                None,
            )
            if table is None:
                raise NoSuchTableError(f"Table '{table_name}' does not exist.")
            schema = CreateTable(table).compile(self.sql_database._engine)
            try:
                for col in schema.statement.columns:
                    s = str(col).split(" ")
                    name = s[0].strip('"')
                    type = s[1]
                    columns.append({"name": name, "type": type})
            except:
                raise
            json_tables[table_name] = columns
        return json_tables

    @staticmethod
    def sources_to_json(sources: list) -> list[dict]:
        if not sources:
            return {}
        source: ToolOutput
        for source in sources:
            if source.raw_output:
                return source.raw_output

    def query(self, query: str) -> QueryReturn:
        return self.query_engine.query(query)

    def find_steps(self, steps: list[StepModel]) -> list[StepModel]:
        """
        Separate the text into a list of logical execution steps for a business user and categorize the step.
        Do not create steps, only refine the given sentences into logical execution steps.
        The steps should be in the order of execution.
        Keep details such as quantities or numbers.
        When choosing a category only use one of the following: [company, people, industry, location, time, campaign, email, extraneous]
        For example if the user query is 'Get me the head of sales for the top 5 companies in the seattle area' the refined query should be:
        step: Get the top 5 companies in the seattle area, category: company
        step: Get the head of sales for each company, category: people

        Args:
            steps (List[StepModel]): A list of StepModel objects, each containing a sentence and a category

        """
        ## Input is not perfect from the llm, it can be a list of steps but in dict form
        ## or it can be a dict with a list of dicts
        ## comes in as a list of steps but in dict form
        if isinstance(steps, dict):
            steps = steps["steps"]
        steps = [StepModel(**step) for step in steps]
        steps = self._refine_step_queries(steps)
        print(f"Returning steps = {steps}")
        return steps

    ## separate company and people steps. Merge filters into the same step.
    def _merge_steps(self, steps: list[StepModel]) -> list[StepModel]:
        """Merge the steps into a list of steps with the same category.

        Args:
            steps (list[dict[str, str]]): The list of steps to merge.

        Returns:
            list[dict[str, str]]: The merged list of steps.
        """
        new_steps: list[StepModel] = []
        old_category = None
        for step in steps:
            if old_category is None or old_category != step.category:
                new_steps.append(step)
            else:
                new_steps[-1].sentence += " " + step.sentence
            old_category = step.category
        return new_steps

    def _fix_proper_nouns(self, sentence: str) -> str:
        """
        Fix the proper nouns in the sentence by replacing them with the correct case from the database.

        Args:
            sentence (str): The sentence to fix.

        Returns:
            str: The fixed sentence.
        """
        ## Replace the potentially wrong case technology name with the correct case from the database
        # sentence = re.sub(tech, proper_name, sentence, re.IGNORECASE)
        return sentence

    def _refine_step_queries(self, steps: list[StepModel]) -> list[StepModel]:
        """
        Refine the step queries by replacing the technology names with the correct case from the database.

        Args:
            steps (list[StepModel]): The list of steps to refine.

        Returns:
            list[StepModel]: The refined list of steps.
        """

        steps = self._merge_steps(steps)
        return steps

    def load_people(self, query: str) -> List[dict]:
        """
        Load people from the database using the query.
        Args:
            query (str): Natural language of the query to load people from the database.
        Returns:
            List[dict[str, str]]: A list of dict, containing the people information.
        """
        if isinstance(query, dict):
            query = query["query"]
        engine = create_engine(self.uri)
        tables = ["Companies", "Industries", "CompanyIndustries", "People"]

        sql_database = SQLDatabase(engine, include_tables=tables)
        query += "return all the people information using '*' in the resulting sql query."
        if self.limit:
            query += f" Place a limit of {self.limit} on the resulting query"

        query_engine = NLSQLTableQueryEngine(
            sql_database=sql_database,
            tables=tables,
        )
        return_dict = []
        r: Response = query_engine.query(str(query))
        if r.metadata and "sql_query" in r.metadata:
            sql_query = r.metadata["sql_query"]
            print("##", sql_query)
            with self.sql_database.engine.connect() as connection:
                result = connection.execute(text(sql_query))
                for item in result.fetchall():
                    return_dict.append({k: v for k, v in zip(item._fields, item._t)})
        return return_dict

    def load_companies(self, query: str) -> List[dict]:
        """
        Load companies from the database using the query.
        Args:
            query (str): Natural language of the query to load companies from the database.
        Returns:
            List[dict]: A list of dict, containing the company information.
        """
        if isinstance(query, dict):
            query = query["query"]
        engine = create_engine(self.uri)
        tables = ["Companies", "Industries", "CompanyIndustries", "People"]
        query += "Return all the company information using '*' in the resulting sql query."
        if self.limit:
            query += f" place a limit of {self.limit} on the resulting query"

        sql_database = SQLDatabase(engine, include_tables=tables)

        query_engine = NLSQLTableQueryEngine(
            sql_database=sql_database,
            tables=tables,
        )
        return_dict = []
        r: Response = query_engine.query(str(query))
        if r.metadata and "sql_query" in r.metadata:
            sql_query = r.metadata["sql_query"]
            print("##", sql_query)
            with self.sql_database.engine.connect() as connection:
                result = connection.execute(text(sql_query))
                for item in result.fetchall():
                    return_dict.append({k: v for k, v in zip(item._fields, item._t)})
        return return_dict

    def get_return_query(self, return_result: QueryReturn) -> str:
        query_list = []
        try:
            source = return_result.sources[-1]
            for raw_output in source.raw_output:
                print("   ###########")
                sm: StepModel = raw_output
                query_list.append(sm.sentence)
            return " ".join(query_list)
        except:
            raise



from typing import Any, Dict, Optional

from llama_index.core.base.llms.types import ChatResponse
from llama_index.core.bridge.pydantic import BaseModel
from llama_index.core.program import LLMTextCompletionProgram


class StructuredResponse:
    model : BaseModel
    response : ChatResponse

class StructuredLLM(LLMTextCompletionProgram):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(
        self,
        llm_kwargs: Optional[Dict[str, Any]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> BaseModel:
        llm_kwargs = llm_kwargs or {}
        if self._llm.metadata.is_chat_model:
            messages = self._prompt.format_messages(llm=self._llm, **kwargs)

            response = self._llm.chat(messages, **llm_kwargs)

            raw_output = response.message.content or ""
        else:
            formatted_prompt = self._prompt.format(llm=self._llm, **kwargs)

            response = self._llm.complete(formatted_prompt, **llm_kwargs)

            raw_output = response.text

        output = self._output_parser.parse(raw_output)
        if not isinstance(output, self._output_cls):
            raise ValueError(
                f"Output parser returned {type(output)} but expected {self._output_cls}"
            )
        return output


import os
from importlib.metadata import version

from dotenv import load_dotenv
from pi_conf import AttrDict, Config, load_config

cfg = load_config("qai")
cfg.to_env()

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

# class QaiFlask(Flask):
#     def __init__(self, *args, cfg=None, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.init(cfg)

#     def init(self, cfg: Config):
#         self.website_2_id: dict[str, str] = None
#         self.id_2_website: dict[str, str] = None
#         self.cfg: Config = cfg

#         if "aws" in self.cfg and self.cfg.aws.get("enabled", True):
#             from qai.storage.aws.aws_history import AWSCompanyIds

#             loader = AWSCompanyIds()
#             self.website_2_id = loader.website_2_id
#             self.id_2_website = loader.id_2_website

#         if "website_name_map" in self.cfg:
#             self.website_2_id = {}
#             self.website_2_id.update(self.cfg.website_name_map)
#             self.id_2_website = {v: k for k, v in self.website_2_id.items()}


# def create_app(_cfg: Config) -> QaiFlask:
#     return QaiFlask(__name__, cfg=_cfg)

try:
    VERSION = version("qai-chat")
except Exception:
    VERSION = "0.0.0"



import os
import time
from functools import partial
from logging import getLogger
from typing import Callable, List, Optional, Self, Union

import requests
from llama_index.agent.openai import OpenAIAgent
from llama_index.core.agent import AgentChatResponse
from llama_index.core.base.llms.types import ChatMessage
from llama_index.core.bridge.pydantic import BaseModel, Field
from llama_index.core.chat_engine.types import AgentChatResponse
from llama_index.core.llms import ChatMessage
from llama_index.core.program import LLMTextCompletionProgram
from llama_index.core.tools import BaseTool
from llama_index.core.tools.tool_spec.base import BaseToolSpec
from llama_index.core.tools.types import BaseTool
from pydantic import BaseModel, Field
from pymongo import MongoClient

from qai.agent import ROOT_DIR, cfg
from qai.agent.utils.distribute import adistribute

log = getLogger(__name__)
PROJ_DIR = os.path.dirname(os.path.dirname(os.path.dirname(ROOT_DIR)))

example_template = """"""
job_title_template = """, who holds the job title {job_title}"""
industry_template = """Industry of {company_name} is {industry}"""
default_template = """
Generate a professional and engaging email from a person named {from_name}{from_job_title_template} from {from_company}, 
addressed to another person named {to_name}{to_job_title_template} from {to_company}.

Rules:

1. Include a subject line and a body.
2. The body should consist of one or two paragraph only, not including tables or making comparisons.
3. The language should be straightforward and devoid of embellishments.
4. Incorporate relevant details about the technologies used by both companies, as well as the industries they operate in, only if it enhances the email’s relevance or context.
5. Format the email to reflect standard business communication practices, including appropriate greetings, paragraph breaks, and closings.

Additional Context:
{from_industry_template}
{to_industry_template}

{example_template}
"""


class EmailModel(BaseModel):
    """
    Email Model for the email tool.
    body is the body of the email with correct punctuation and newlines and spacing.
    """

    subject: str = Field(description="The subject of the email.")
    body_list: list[str] = Field(
        description="The body of the email as a list of str. There can be empty lines in the list."
    )
    # body: str = Field(description="The properly formatted body of the email.")
    email: Optional[str] = Field(description="The email address of the person.")
    name: Optional[str] = Field(description="The name of the person.")
    phone: Optional[str] = Field(description="The phone number of the person.")

    @property
    def body(self) -> str:
        return "\n".join(self.body_list)

    def dict(self):
        return {"subject": self.subject, "body": self.body, "email": self.email}


class EmailToolSpec(BaseToolSpec):
    """Tool for emails."""

    template: str = None
    spec_functions = ["draft_email"]
    from_person: dict
    from_company: dict

    def __init__(
        self,
        from_person: dict,
        from_company: dict,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.from_person = from_person
        self.from_company = from_company

    def create_prompt(
        self,
        from_person: dict,
        from_company: dict,
        to_person: dict,
        to_company: dict,
    ) -> str:
        """
        Create a prompt for the email.
        Args:
            from_person (dict): The person sending the email.
            from_company (dict): The company sending the email.
            to_person (dict): The person receiving the email.
            to_company (dict): The company receiving the email.

        """
        from_name = self.get_name(from_person)
        to_name = self.get_name(to_person)

        from_company_name = self.get_cname(from_company)
        to_company_name = self.get_cname(to_company)

        from_job_title = from_person.get("job_title")
        to_job_title = to_person.get("job_title")
        from_job_title_template = (
            "" if not from_job_title else job_title_template.format(job_title=from_job_title)
        )
        to_job_title_template = (
            "" if not to_job_title else job_title_template.format(job_title=to_job_title)
        )
        from_company_industry = from_company.get("industry")
        to_company_industry = to_company.get("industry")

        from_industry_template = ""
        if from_company_industry:
            from_industry_template = industry_template.format(
                company_name=from_company_name, industry=from_company_industry
            )
        to_industry_template = ""
        if to_company_industry:
            to_industry_template = industry_template.format(
                company_name=to_company_name, industry=to_company_industry
            )

        s = default_template.format(
            from_name=from_name,
            from_company=from_company_name,
            to_name=to_name,
            to_company=to_company_name,
            from_job_title_template=from_job_title_template,
            to_job_title_template=to_job_title_template,
            from_industry_template=from_industry_template,
            to_industry_template=to_industry_template,
            example_template=example_template,
        )
        return s

    def get_cname(self, company: dict) -> str:
        """
        Get the company name from the company dict.
        """
        cname = company.get("company_name")
        if not cname:
            cname = company.get("name")
        if not cname:
            raise ValueError(f"Company must have a name. values={company}")
        return cname

    def get_name(self, person: dict) -> str:
        """
        Get the name from the person dict.
        """
        name = person.get("name")
        if not name:
            first_name = person.get("first_name")
            last_name = person.get("last_name")
            if first_name and last_name:
                name = f"{first_name} {last_name}"
            else:
                name = first_name or last_name
        if not name:
            raise ValueError(f"Person must have a name. values={person}")
        return name

    def draft_email(
        self,
        to_person: dict,
        to_company: dict,
        prompt: Optional[str] = None,
        callback: Optional[Callable] = None,
    ) -> EmailModel:
        """
        Draft an email for a person.
        """
        prompt = prompt or self.create_prompt(
            self.from_person,
            self.from_company,
            to_person,
            to_company,
        )

        # email = EmailModel(subject=data["subject"], body=data["body"])
        ## TODO - Email body is not being generated with correct formatting when using json output.
        ## using a default fixes it but the call needs to be manually parsed.
        program = LLMTextCompletionProgram.from_defaults(
            output_cls=EmailModel, prompt_template_str=prompt, verbose=True
        )
        email: EmailModel = program()
        email.email = to_person.get("email")
        email.name = to_person.get("name")
        email.phone = to_person.get("phone")
        if callback:
            callback(email)
        return email

    @staticmethod
    def _create_email(
        from_person: dict,
        from_company: dict,
        to_person: dict,
        to_company: dict,
        template: Optional[str] = None,
        email_callback: Optional[Callable] = None,
    ) -> EmailModel:
        log.debug(f"Creating email from {from_person} to {to_person}")
        log.debug(f"  From Company: {from_company} To Company: {to_company}")
        company_email = to_person.get("email")
        if company_email:
            company_email = company_email.split("@")[1]
        to_company = {
            "name": company_email,
        }
        if not to_company:
            raise ValueError(f"Company not found for person {to_person}")
        draft_tool = EmailToolSpec(from_person=from_person, from_company=from_company)
        email = draft_tool.draft_email(
            to_person=to_person,
            to_company=to_company,
            callback=email_callback,
        )

        return email

    @staticmethod
    def _create_emails(
        kwargs_list: list[dict],
        on_complete: Optional[Callable] = None,
        asynchronous: bool = True,
    ) -> None:
        if asynchronous:
            adistribute(
                EmailToolSpec._create_email,
                kwargs_list=kwargs_list,
                on_complete=on_complete,
                nprocs=4,
            )
        else:
            for kwargs in kwargs_list:
                EmailToolSpec._create_email(**kwargs)
            if on_complete:
                on_complete()


class EmailAgent(OpenAIAgent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def chat(
        self,
        message: str,
        chat_history: Optional[List[ChatMessage]] = None,
        tool_choice: Optional[Union[str, dict]] = None,
    ) -> AgentChatResponse:
        return super().chat(message, chat_history, tool_choice)

    @staticmethod
    def on_email_complete(sequence_id, email: EmailModel) -> None:
        print(f"Email completed: {sequence_id} {email}")
        mongo = cfg.mongo
        client = MongoClient(mongo.uri)
        db = client[mongo.db]

        collection = db[mongo.collection]
        js = {
            "sequence_id": sequence_id,
            "prospect_email": email.email,
            "prospect_name": email.name,
            "prospect_phone": email.phone,
            "message_subject": email.subject,
            "message_body": email.body,
            "is_message_generation_complete": True,
        }
        print(f"Inserting into collection: {mongo.uri} {db} {collection.name} ")
        result = collection.insert_one(js)
        print(f"Insert Result: {result}")

    @staticmethod
    def on_all_emails_complete(sequence_id: str, successes: int, errors: int) -> None:
        print(f"All Email completed: {successes} success, {errors} errors.")
        url = cfg.email.on_complete_emails_url
        print(f"Sending call to {url}  sequence_id={sequence_id}")
        delay = cfg.email.get("delay_complete_request", 0)
        if delay:
            time.sleep(delay)
        r = requests.post(url, json={"sequence_id": sequence_id})
        print(r.json())

    def create_emails(
        self,
        sequence_id: str,
        from_person: dict,
        from_company: dict,
        to_people: dict,
        to_companies: dict = None,
        use_async_on: int = 1,
    ) -> List[EmailModel]:
        """
        Create the emails for the campaign.
        Args:
            None
        Returns:
            A list of EmailModel objects.
        """
        log.debug("Sending emails")
        if to_companies is None:
            to_companies = {}
        email_params = []
        async_emails_params = []
        ### Resolve the company for each person

        for i, person_id in enumerate(to_people.keys()):
            to_person = to_people[person_id]
            if "company_id" in to_person:
                to_company = to_companies.get(to_person["company_id"])
            try:
                if not to_company:
                    to_company = {"name": to_person.get("email").split("@")[1]}
            except:
                to_company = {
                    "name": "Unknown",
                }
            params = {
                "from_person": from_person,
                "from_company": from_company,
                "to_person": to_person,
                "to_company": to_company,
            }
            log.debug(
                f"Creating email for {to_person} to {to_company} from {from_person} {from_company}"
            )

            if i < use_async_on:
                email_params.append(params)
            else:
                f = partial(EmailAgent.on_email_complete, sequence_id)
                params["email_callback"] = f
                async_emails_params.append(params)
        if async_emails_params:
            f = partial(EmailAgent.on_all_emails_complete, sequence_id)
            ## TODO resolve potential issue of async completing before
            # all sync emails are generated. and sending message before
            EmailToolSpec._create_emails(
                kwargs_list=async_emails_params,
                on_complete=f,
                asynchronous=True,
            )
        emails = []
        for ep in email_params:
            email = EmailToolSpec._create_email(**ep)
            EmailAgent.on_email_complete(sequence_id, email)
            emails.append(email)
            if not async_emails_params:
                EmailAgent.on_all_emails_complete(sequence_id, 1, 0)
        return emails

    @staticmethod
    def create(
        from_person: dict,
        from_company: dict = None,
        tools: List[BaseTool] = None,
    ) -> Self:
        email_tool = EmailToolSpec(from_person, from_company)
        tools = tools or email_tool.to_tool_list()
        agent = EmailAgent.from_tools(tools)
        return agent



from enum import StrEnum
from typing import Any, Self, Sequence

from llama_index.agent.openai import OpenAIAgent
from llama_index.core.bridge.pydantic import BaseModel, Field
from llama_index.core.llms.llm import LLM
from llama_index.core.tools import BaseTool, FunctionTool

# from .tools.types import ToolMetadata
from llama_index.core.tools.types import ToolMetadata
from pydantic import BaseModel, Field


class AgentType(StrEnum):
    campaign = "campaign"
    database = "database"
    conversation = "conversation"
    website_navigator = "website_navigator"
    # new_workflow = "new_workflow"
    # activate_existing_workflow = "activate_existing_workflow"

    def __eq__(self, other: str | Self) -> bool:
        try:
            return super().__eq__(other)
        except:
            if str(self.value) == str(other):
                return True
            raise


class FindAgentModel(BaseModel):
    sentence: str = Field(description="The user sentence to find the correct agent for.")
    agent_type: AgentType = Field(description="Which agent the user action belongs to.")


def find_agent(sentence: str, agent_type: AgentType) -> tuple[str, AgentType]:
    print(f"Found agent for sentence: {sentence} agent_type: {agent_type}")

    return sentence, agent_type


class FindAgentWorker(OpenAIAgent):
    """An Agent to find the best agent given a user sentence."""

    name: str = "find_agent"
    description: str = (
        f"Find the correct agent given a user sentence. "
        f"Only use one of the following agent types: [{','.join([str(e) for e in AgentType])}]"
    )

    @classmethod
    def create(cls, nametools: Sequence[BaseTool] = None, llm: LLM = None, *args, **kwargs: Any):
        print("#######")
        print(str(cls.name))
        print("#######")
        find_agent_tool = FunctionTool.from_defaults(
            name=cls.name,
            fn=find_agent,
            # description=description,
            # fn_schema=FindAgentModel,
            tool_metadata=ToolMetadata(
                name=cls.name,
                description=cls.description,
                fn_schema=FindAgentModel,
            ),
        )
        agent = FindAgentWorker.from_tools(llm=llm, tools=[find_agent_tool], verbose=True)

        return agent


def test_find_agent():
    from pi_conf import load_config

    cfg = load_config("qai-chat")
    cfg.to_env()
    agent = FindAgentWorker.create()

    sentence = "I want to create a campaign"
    # llm = OpenAI(model="gpt-3.5-turbo-0125")
    # from llama_index.core.tools import BaseTool, FunctionTool

    # name = "find_agent"
    # description = (
    #     f"Find the correct agent given a user sentence. "
    #     f"Only use one of the following agent types: [{','.join([str(e) for e in AgentType])}]"
    # )
    # find_agent_tool = FunctionTool.from_defaults(
    #     name=name,
    #     fn=find_agent,
    #     # description=description,
    #     # fn_schema=FindAgentModel,
    #     tool_metadata=ToolMetadata(
    #         name=name,
    #         description=description,
    #         fn_schema=FindAgentModel,
    #     ),
    # )
    # agent = OpenAIAgent.from_tools(llm=llm, tools=[find_agent_tool], verbose=True)
    # # print(agent.chat(sentence, tool_choice="find_agent"))
    print(agent.chat(sentence))
    print("FINISHED")


if __name__ == "__main__":
    test_find_agent()



from enum import StrEnum


class StringEnum(StrEnum):
    """
    A StringEnum is a StrEnum that can be compared to a string.
    """

    def __eq__(self, other: str | object) -> bool:
        """
        Compare the StringEnum to the __str__ of an object."""
        try:
            return super().__eq__(other)
        except:
            if str(self.value) == str(other):
                return True
            raise




from qai.agent.data.state_codes import state_codes


def refine_step_queries(steps: list[dict[str, str]]) -> list[dict[str, str]]:
    """
    Refine the step queries by replacing the technology names with the correct case from the database.

    Args:
        steps (list[dict[str, str]]): The list of steps to refine.

    Returns:
        list[dict[str, str]]: The refined list of steps.
    """

    ## separate company and people steps. Merge filters into the same step.
    def merge_steps(steps: list[dict[str, str]]) -> list[dict[str, str]]:
        """Merge the steps into a list of steps with the same category.

        Args:
            steps (list[dict[str, str]]): The list of steps to merge.

        Returns:
            list[dict[str, str]]: The merged list of steps.
        """
        new_steps = []
        old_step = None
        for d in steps:
            cat, s = d["category"], d["sentence"]
            is_company_step = cat.startswith("company")
            is_person_step = cat.startswith("people")
            is_industry_step = cat.startswith("industr")
            if is_company_step:
                if old_step == is_company_step:
                    new_steps[-1][1] += " " + s
                else:
                    new_steps.append(["company", s])
                old_step = is_company_step
            elif is_industry_step:
                if old_step == is_industry_step:
                    new_steps[-1][1] += " " + s
                else:
                    new_steps.append(["industry", s])
                old_step = is_industry_step
            elif is_person_step:
                if old_step == is_person_step:
                    new_steps[-1][1] += " " + s
                else:
                    new_steps.append(["people", s])
                old_step = is_person_step
            else:
                new_steps.append([cat, s])
                old_step = None
        return new_steps

    steps = merge_steps(steps)

    def replace(sentence: str) -> str:
        # for tech, proper_name in technologies.items():
        #     ## Replace the potentially wrong case technology name with the correct case from the database
        #     sentence = re.sub(tech, proper_name, sentence, re.IGNORECASE)
        return sentence

    refined_steps = []
    for cat, sentence in steps:
        refined_steps.append([cat, replace(sentence)])
    return refined_steps


def modify_sql_query(original_sql, prefix=""):
    """
    Modifies the provided SQL query by replacing the initial SELECT statement
    with 'SELECT id'. This function assumes the query contains joins from multiple tables.

    Args:
    original_sql (str): The original SQL query string.

    Returns:
    str: The modified SQL query with 'SELECT id' as the initial statement.
    """

    # Split the query into segments
    segments = original_sql.split("FROM")

    # Check if the query is valid (has at least one FROM segment)
    if len(segments) < 2:
        raise ValueError("The SQL query does not seem to contain a valid 'FROM' clause.")
    seg0 = segments[0]
    if not prefix and "." in seg0:
        ## Find the prefix after the SELECT
        prefix = seg0.split("SELECT")[1].split(".")[0] + "."

    # Replace the initial SELECT statement with 'SELECT id'
    segments[0] = f"SELECT {prefix}id, * "

    # Reassemble the query
    modified_sql = "FROM".join(segments)

    ## State code check
    if "location_state" in modified_sql:
        for state_longname, state_code in state_codes.items():
            modified_sql = modified_sql.replace(state_longname, state_code)

    return modified_sql



import uuid
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Self, Union

from llama_index.agent.openai import OpenAIAgent
from llama_index.core.agent import AgentChatResponse, Task
from llama_index.core.agent.types import Task
from llama_index.core.base.llms.types import ChatMessage
from llama_index.core.bridge.pydantic import BaseModel, Field
from llama_index.core.chat_engine.types import AgentChatResponse
from llama_index.core.llms import ChatMessage
from llama_index.core.tools import BaseTool
from llama_index.core.tools.tool_spec.base import BaseToolSpec
from llama_index.core.tools.types import BaseTool

from qai.agent import cfg
from qai.agent.agents.email import EmailAgent, EmailModel
from qai.agent.agents.sql_query import RefineSQLQuery, StepModel
from qai.agent.models import OutreachType


class PersonID(BaseModel):
    id: str = Field(description="The database id of the person.")


class PersonModel(BaseModel):
    id: Optional[str] = Field(description="The database id of the person.")
    name: Optional[str] = Field(description="The name of the person.")
    email: Optional[str] = Field(description="The email of the person.")
    phone: Optional[str] = Field(description="The phone number of the person.")
    linkedin: Optional[str] = Field(description="The linkedin profile of the person.")
    whatsapp: Optional[str] = Field(description="The whatsapp number of the person.")


class CompanyCampaignModel(BaseModel):
    people: list[str] = Field(description="A list of people to generate a campaign for.")
    companies: list[str] = Field(description="A list of companies to generate a campaign for.")
    outreach_types: list[OutreachType] = Field(
        description="The type of outreach to use for the campaign."
    )


class CampaignToolSpec(BaseToolSpec):
    agent: "CampaignAgent" = None
    companies: list[str] = Field(description="A list of companies to generate a campaign for.")
    spec_functions = ["create_campaign", "text_to_outreach_type"]

    def __init__(self, *args, **kwargs):
        pass

    def text_to_outreach_type(self, outreach_types: list[OutreachType]) -> list[OutreachType]:
        """
        Convert a natural langague of strings to a list of OutreachType objects.
        Args:
            outreach_types: A list of strings representing the outreach types.
        Returns:
            A list of OutreachType objects.
        """
        print(f"Converting outreach types: {outreach_types}")
        return outreach_types

    def create_campaign(
        self,
        people: list[PersonID],
        outreach_types: list[OutreachType],
    ) -> str:
        """
        Create a campaign for a list of PersonID objects. These id's should never be inferred.
        If they are missing, ask for them.
        Args:
            people: A list of PersonID objects to generate a campaign for.
            outreach_types: A list of OutreachType objects for how to contact the people.
        Returns:
            A string indicating the campaign was created.
        """
        # people = self.people
        print(
            f"Creating campaign for # people: {len(people)}  # outreach_types: {len(outreach_types)}, people={people}"
        )
        for p in people:
            print(f"Person: {p} Outreach: {outreach_types}")

        return "Campaign is starting to be created, it can be found in the campaign tab."


@dataclass
class CampaignResponse(AgentChatResponse):
    completed: bool = False
    emails: List[EmailModel] = None


class CampaignAgent(OpenAIAgent):
    """
    An agent that generates a campaign for a list of people and companies.
    """

    def __init__(self, tools: List[BaseTool], **kwargs: Any) -> None:
        """Init params."""
        super().__init__(
            tools=tools,
            **kwargs,
        )
        # The person to send the emails from.
        self.from_person: dict = {}
        # The company to send the emails from.
        self.from_company: dict = {}
        # A dict of people to generate a campaign for.
        self.people: dict = {}
        # A dict of companies to generate a campaign for.
        self.companies: dict = {}
        # The type of outreach to use for the campaign.
        self.outreach_types: list[OutreachType] = [OutreachType.email]
        # The URL to call when the emails are completed.
        self.on_complete_emails_url: Optional[str] = None

        self.email_agent: EmailAgent = None

    @staticmethod
    def _email_on_complete(url: str = None) -> None:
        print(f"Call sent to {url}")

    def _initialize_state(self, task: Task, **kwargs: Any) -> Dict[str, Any]:
        """Initialize state."""
        print(f"Initializing state for task {task} with kwargs {kwargs}")
        return {"count": 0, "current_reasoning": []}

    def chat(
        self,
        message: str,
        chat_history: Optional[List[ChatMessage]] = None,
        tool_choice: Optional[Union[str, dict]] = None,
    ) -> CampaignResponse:
        """
        Chat with the agent.
        """
        print(f"Chatting with message {message} and tool_choice {tool_choice}")
        print(
            f"  People IDs: {len(self.people)}, company IDs: {len(self.companies)}, "
            f"outreach types: {len(self.outreach_types)}"
        )

        refine_tools = RefineSQLQuery(**cfg.db)

        refining_agent = OpenAIAgent.from_tools(
            refine_tools.to_tool_list(),
            verbose=True,
            system_prompt=refine_tools.refine_query_system_message,
        )
        r: AgentChatResponse = refining_agent.chat(message, tool_choice="find_steps")
        if r.sources and r.sources[0].tool_name == "find_steps":
            tool = r.sources[0]
            steps: list[StepModel] = tool.raw_output
            print(f"  Tool output: ")
            for step in steps:
                try:
                    if step.category == "people":
                        result = refine_tools.load_people(step.sentence)
                        ## result is a list of dict, change to dict
                        if result:
                            result = {r["id"]: r for r in result}
                        self.people.update(result)
                        print(f"  Added People: {len(result)}")
                    elif step.category == "company":
                        result = refine_tools.load_companies(step.sentence)
                        ## result is a list of dict, change to dict
                        if result:
                            result = {r["id"]: r for r in result}
                        self.companies.update(result)
                        print(f"  Added Companies: {len(result)}")

                    else:
                        result = "No Result"
                except Exception as e:
                    print(f"Error: {e}\nStep={step}\nResult: {result}")
                    raise
                print(f"  Result: {result}")
        npeople = len(self.people) > 0
        ncompanies = len(self.companies) > 0
        noutreach = len(self.outreach_types) > 0
        assistant_message = ". I have the following:"
        if npeople:
            assistant_message += f"\n  People: {list(self.people.keys())[0]}"
        if ncompanies:
            assistant_message += f"\n  Companies: {list(self.companies.keys())[0]}"
        if noutreach:
            assistant_message += (
                f"\n  Outreach types: {' '.join([str(o) for o in self.outreach_types])}"
            )
        ## I am missing the following
        if not npeople or not noutreach:
            assistant_message += "\n. I am missing the following information :"
            if not npeople:
                assistant_message += "\n    People"
            ## Purposefully not using companies as we don't really need them
            if not noutreach:
                assistant_message += "\n    Outreach types"
        message += assistant_message
        response: AgentChatResponse = super().chat(message, chat_history, tool_choice)
        completed = response.sources and response.sources[0].tool_name == "create_campaign"
        setattr(response, "completed", completed)
        emails = None
        sequence_id = str(uuid.uuid4())

        if completed:
            emails = self.email_agent.create_emails(
                sequence_id, self.from_person, self.from_company, self.people, self.companies
            )
        setattr(response, "emails", emails)
        print(f"campaignagent::Chat completed={completed}  response: {response} emails: {emails}")
        return response

    def _finalize_task(self, task: Task, **kwargs: Any) -> None:
        """Finalize task."""
        print(f"Finalizing task with state {task} and kwargs {kwargs}")
        # nothing to finalize here

    @staticmethod
    def create(
        people: list[dict] = None,
        companies: list[dict] = None,
        outreach_types: list[OutreachType] = None,
        on_complete_emails_url: str = None,
        tools: List[BaseTool] = None,
        verbose: bool = False,
        from_company: dict = None,
        from_person: dict = None,
        model_config: dict = None,
        system_prompt: str = None,
        email_agent: EmailAgent = None,
        **kwargs,
    ) -> Self:
        campaign_tool = CampaignToolSpec()
        tools = tools or campaign_tool.to_tool_list()
        system_prompt = system_prompt or (
            "Create a campaign for a list of people and companies. If a parameter is missing,"
            " please ask for it, do not try to infer it."
        )
        agent: CampaignAgent = CampaignAgent.from_tools(
            tools=tools,
            system_prompt=system_prompt,
            verbose=verbose,
            **kwargs,
        )
        agent.people = people or {}
        agent.companies = companies or {}
        agent.on_complete_emails_url = on_complete_emails_url
        agent.from_person = from_person or {}
        agent.from_company = from_company or {}
        agent.email_agent = email_agent or EmailAgent.create(
            agent.from_person,
            agent.from_company,
        )
        return agent




from llama_index.core import Settings
from llama_index.core.bridge.pydantic import BaseModel, Field
from llama_index.core.settings import Settings
from llama_index.core.tools import FunctionTool
from llama_index.core.tools.tool_spec.base import BaseToolSpec
from llama_index.core.tools.types import ToolMetadata, ToolOutput
from llama_index.llms.openai import OpenAI
from pi_conf import load_config

from qai.agent.agents.find_agent import AgentType
from qai.agent.tools.types import StringEnum

Settings.llm = OpenAI(model="gpt-3.5-turbo", temperature=0.1)


class QueryType(StringEnum):
    campaign = "campaign"
    company = "company"
    people = "people"


class QueryTypeModel(BaseModel):
    sentence: str = Field(description="The user sentence to find the correct query type for.")
    query_type: QueryType = Field(description="Which query type sentence belongs to.")


def find_query_type(sentence: str, query_type: QueryType) -> AgentType:
    """
    Find the correct query type given a user sentence.
    Only use one of the following query types: [campaign, company, people]
    Args:
        sentence: The user sentence to find the correct query type for.
        query_type: Which query type the user action belongs to.
    Returns:
        The query type.
    """
    print(f"Found find_query_type sentence: {sentence} agent_type: {query_type}")
    if isinstance(query_type, dict):
        query_type = query_type["query_type"]
    return query_type


class BasicQuery(BaseToolSpec):
    """Simple query tool that returns the query type of the user sentence."""

    step_functions: list[str] = ["find_query_type"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.find_query_tool = FunctionTool.from_defaults(
            name="find_query_type",
            fn=find_query_type,
            tool_metadata=ToolMetadata(
                name="find_query_type",
                description="Find the correct query type given a user sentence. Only use one of the following query types: [campaign, company, people]",
                fn_schema=QueryTypeModel,
            ),
        )

    @staticmethod
    def sources_to_json(sources: list) -> list[dict]:
        if not sources:
            return {}
        source: ToolOutput
        for source in sources:
            if source.raw_output:
                return source.raw_output

    def query(self, sentence: str) -> QueryType:
        """Query the user sentence to find the correct query type."""

        # basic_agent = OpenAIAgent.from_tools([self.find_query_tool])
        # r = basic_agent.chat(sentence)
        # return r.sources[0]
        sentence = sentence.lower()
        if "compan" in sentence:
            return QueryType.company
        elif "campaign" in sentence:
            return QueryType.campaign
        elif "people" in sentence:
            return QueryType.people
        else:
            return QueryType.campaign


if __name__ == "__main__":
    cfg = load_config("qai")
    basic_query = BasicQuery()
    query = "What is the best way to reach out to people?"
    r = basic_query.query(query)
    print(type(r))



from enum import StrEnum
from typing import Self


class Step(StrEnum):
    create_campaign = "create_campaign"
    company_list = "company_list"
    company_filter = "company_filter"
    company_more_info = "company_more_info"
    people_list = "people_list"
    people_filter = "people_filter"
    people_more_info = "people_more_info"
    industry_list = "industry_list"
    industry_filter = "industry_filter"
    industry_more_info = "industry_more_info"
    action_email = "action_email"
    action_sms = "action_sms"
    action_call = "action_call"
    action_reminder = "action_reminder"
    action_graph_results = "action_graph_results"
    action_table_results = "action_table_results"
    more_info = "more_info"
    extraneous = "extraneous"

class Action(StrEnum):
    create_campaign = "create_campaign"
    database_query = "database_query"
    conversation = "conversation"
    goto_website_location = "goto_website_location"
    new_workflow = "new_workflow"
    activate_existing_workflow = "activate_existing_workflow"
    
    def __eq__(self, other: str | Self) -> bool:
        try:
            return super().__eq__(other)
        except:
            if str(self.value) == str(other):
                return True
            raise



print([str(a) for a in Action])
fn_break_into_steps = {
    "type": "function",
    "function": {
        "name": "break_into_steps",
        "description": (
            "Separate the text into a list of logical execution steps for a business user. "
            "The steps should be in the order of execution. "
            "Keep details such as quantities or numbers."
            "Non business steps should be classified as extraneous. "
            f"Categories should only be one of [{','.join([str(e) for e in Step])}]"
        ),
        "parameters": {
            "type": "object",
            "properties": {
                "steps": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "sentence": {
                                "type": "string",
                                "description": (
                                    "A revised sentence that can be used for the steps, "
                                    "the revised sentence keeps details such as numbers."
                                ),
                            },
                            "category": {
                                "type": "string",
                                "enum": [str(e) for e in Step],
                                "description": "A category for the revised sentence. Describes whether the sentence is about companies, people, industries, or an action.",
                            },
                        },
                    },
                    "description": "The sentence that describes the step",
                },
            },
            "required": ["steps", "sentence", "category"],
        },
    },
}

fn_find_action = {
    "type": "function",
    "function": {
        "name": "find_action",
        "description": (("Determine the action that the user wants to take. Always respond with values for all parameters in this tool.")),
        "parameters": {
            "type": "object",
            "properties": {
                "sentence": {
                    "type": "string",
                    "description": "A revised sentence that is what the user wants to do.",
                },
                "category": {
                    "type": "string",
                    "enum": [str(e) for e in Action],
                    "description": "Which category the user action belongs to.",
                },
            },
            "required": ["sentence", "category"],
        },
    },
}


# break_into_steps = {
#     "steps": [
#         {
#             "category": "company_filter",
#             "sentence": "Find companies that use Chili Piper.",
#         },
#         {
#             "category": "company_filter",
#             "sentence": "Find companies that use Drift.",
#         },
#         {
#             "category": "company_filter",
#             "sentence": "Filter the companies to only include those that use "
#             "both Chili Piper and Drift.",
#         },
#         {
#             "category": "action_email",
#             "sentence": "Send an email to each of the filtered companies.",
#         },
#     ]
# }



import atexit
import multiprocessing
import threading
import traceback
from logging import getLogger
from queue import Queue
from typing import Any, Callable

log = getLogger(__name__)
pool_id = 0

pool_handlers: dict[int, "PoolHandler"] = {}


class Worker:
    """
    A worker class that can be used to distribute work across multiple processes.
    """
    def __init__(
        self,
        func: Callable,
        results_queue: Queue,
        error_queue: Queue,
        nexpected_results: int = None,
        on_complete: Callable = None,
    ):
        self.func = func
        self.results_queue = results_queue
        self.error_queue = error_queue

        self.on_complete = on_complete
        self.nexpected_results = nexpected_results

    def _on_complete(self):
        try:
            self.on_complete(self.results_queue.qsize(), self.error_queue.qsize())
        except Exception as e:
            log.error(f"Error in on_complete: {e}")
            self.on_complete(-1, 1)

    def __call__(self, *args, **kwargs):
        try:
            log.debug(f"Worker: Processing {self.func} args={args} with kwargs={kwargs}")
            result = self.func(*args, **kwargs)

            self.results_queue.put(result)
            return result
        except Exception as e:
            # Send error information to the main process via error_queue
            log.error(traceback.format_exc())
            log.error(f"Worker: Error Processing {self.func} args={args} with kwargs={kwargs}")

            self.error_queue.put(e)
            return None
        finally:
            # print(len(self.successes) + len(self.errors), self.nexpected_results)
            if self.on_complete:
                ncompleted = self.results_queue.qsize() + self.error_queue.qsize()

                if self.on_complete and self.nexpected_results == ncompleted:
                    self._on_complete()


class PoolHandler:
    """
    A PoolHandler class that can be used to manage a pool of processes.
    """
    def __init__(
        self,
        pool_id: int,
        pool: multiprocessing.Pool,
        results_queue: Queue,
        error_queue: Queue,
        manager: multiprocessing.Manager,
    ):
        self.pool_id = pool_id
        self.pool = pool
        self.results_queue = results_queue
        self.error_queue = error_queue
        self.manager = manager
        self.results = []
        self.errors = []

    def close(self):
        self.pool.close()
        self.pool.join()
        pool_handlers.pop(self.pool_id, None)

    def _get_results(self):
        while not self.results_queue.empty():
            a = self.results_queue.get()
            self.results.append(a)
        while not self.error_queue.empty():
            self.errors.append(self.error_queue.get())
        return self.results, self.errors

    def get_results(self):
        return self._get_results()[0]

    def get_errors(self):
        return self._get_results()[1]

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def distribute(
    func,
    args_list: list[Any] = None,
    kwargs_list: list[Any] = None,
    nprocs: int = None,
    callback: Callable = None,
    error_callback: Callable = None,
    on_complete: Callable = None,
):
    """
    Distribute a function across multiple processes.
    Args:
        func (Callable): The function to distribute.
        args_list (list): A list of arguments to pass to the function.
        kwargs_list (list): A list of keyword arguments to pass to the function.
        nprocs (int): The number of processes to use.
        callback (Callable): A callback function to call when the function completes.
        error_callback (Callable): A callback function to call when an error occurs.
        on_complete (Callable): A callback function to call when all functions have completed.
            on_complete(successes: List, errors: List)
    """
    pool_handler = _distribute(
        func,
        args_list=args_list,
        kwargs_list=kwargs_list,
        nprocs=nprocs,
        callback=callback,
        error_callback=error_callback,
        on_complete=on_complete,
    )
    # Close the pool and wait for work to finish
    pool_handler.close()
    results = pool_handler.get_results()
    return results


def adistribute(
    func,
    args_list: list[Any] = None,
    kwargs_list: list[Any] = None,
    nprocs: int = None,
    callback: Callable = None,
    error_callback: Callable = None,
    on_complete: Callable = None,
) -> PoolHandler:
    """
    Start a distribution across multiple processes and return immediately.
    Args:
        func (Callable): The function to distribute.
        args_list (list): A list of arguments to pass to the function.
        kwargs_list (list): A list of keyword arguments to pass to the function.
        nprocs (int): The number of processes to use.
        callback (Callable): A callback function to call when the function completes.
        error_callback (Callable): A callback function to call when an error occurs.
        on_complete (Callable): A callback function to call when all functions have completed.
            on_complete(successes: List, errors: List)
    """
    pool_handler = _distribute(
        func,
        args_list=args_list,
        kwargs_list=kwargs_list,
        nprocs=nprocs,
        callback=callback,
        error_callback=error_callback,
        on_complete=on_complete,
    )

    with threading.Lock():
        global pool_id
        pool_handlers[pool_handler.pool_id] = pool_handler
        pool_id += 1
    return pool_handler


def _distribute(
    func,
    args_list: list[Any] = None,
    kwargs_list: list[Any] = None,
    nprocs: int = None,
    callback: Callable = None,
    error_callback: Callable = None,
    on_complete: Callable = None,
) -> PoolHandler:
    """
    Distribute a function across multiple processes.
    Args:
        func (Callable): The function to distribute.
        args_list (list): A list of arguments to pass to the function.
        kwargs_list (list): A list of keyword arguments to pass to the function.
        nprocs (int): The number of processes to use.
        callback (Callable): A callback function to call when the function completes.
        error_callback (Callable): A callback function to call when an error occurs.
        on_complete (Callable): A callback function to call when all functions have completed.
            on_complete(successes: List, errors: List)
    """
    # Creating a Queue
    manager = multiprocessing.Manager()

    nprocs = nprocs or multiprocessing.cpu_count()
    if args_list is None and kwargs_list is None:
        raise ValueError("args_list and kwargs_list cannot both be None")
    size_args = len(args_list) if args_list else 0
    size_kwargs = len(kwargs_list) if kwargs_list else 0
    size = max(size_args, size_kwargs)
    if not args_list:
        args_list = [None] * size
    if kwargs_list is None:
        kwargs_list = [{}] * size  # Empty dictionaries for kwargs if none provided

    ## Create a pool of processes
    pool = multiprocessing.Pool(processes=nprocs)
    results_queue = manager.Queue()
    error_queue = manager.Queue()

    worker = Worker(
        func=func,
        results_queue=results_queue,
        error_queue=error_queue,
        nexpected_results=size,
        on_complete=on_complete,
    )

    ## Distribute the work
    for i in range(size):
        args = (args_list[i],) if i < size_args else ()
        kwargs = kwargs_list[i] if i < size_kwargs else {}
        pool.apply_async(
            worker,
            args=args,
            kwds=kwargs,
            callback=callback,
            error_callback=error_callback,
        )
    ph = PoolHandler(pool_id, pool, results_queue, error_queue, manager)
    return ph


def wait_all():
    """Wait for all pool handlers to finish and close."""
    for k in list(pool_handlers.keys()):
        ph = pool_handlers.pop(k, None)
        try:
            ph.close()
        except Exception as e:
            log.error(f"Error closing pool handler: {e}")


atexit.register(wait_all)



import sqlite3
from urllib.request import pathname2url


def verify_db_connection(connection_str: str):
    ## Verify the database connection
    try:
        uri = connection_str.replace("sqlite:///", "")
        dburi = "file:{}?mode=rw".format(pathname2url(uri))
        sqlite3.connect(dburi, uri=True)
    except sqlite3.OperationalError:
        # handle missing database case
        raise



from dataclasses import dataclass, field

from llama_index.agent.openai import OpenAIAgent
from llama_index.core.base.llms.types import ChatMessage, MessageRole
from llama_index.core.memory.types import BaseMemory
from qai.ai import MessageRole
from qai.ai.frameworks.openai.llm import OpenAILLM

from qai.agent.agents.find_agent import AgentType

tables = ["Companies", "Industries", "CompanyIndustries"]


@dataclass
class QaiSession(OpenAIAgent):
    db_config: dict
    model_config: dict
    session_id: str = None
    agent_type: AgentType = None
    messages: list[dict] = field(default_factory=list)
    user_id: str = None
    memory: BaseMemory = None

    def chat(self, query: str):
        self.memory.chat_store.add_message(
            key=self.user_id,
            message=ChatMessage(
                content=query,
                role=MessageRole.USER,
            ),
        )
        msgs = self.memory.chat_store.get_messages(self.user_id)
        msgs = [{"role": msg.role, "content": msg.content} for msg in msgs]
        llm = OpenAILLM(config=self.model_config)
        qr = llm.query(query, messages=msgs)

        self.memory.chat_store.add_message(
            key=self.user_id,
            message=ChatMessage(**qr.to_message()),
        )



from typing import Any, Callable, Dict, List, Optional, Sequence, Type

from llama_index.agent.openai import OpenAIAgent
from llama_index.agent.openai.base import DEFAULT_MAX_FUNCTION_CALLS
from llama_index.core.base.llms.types import ChatMessage
from llama_index.core.bridge.pydantic import BaseModel, Field
from llama_index.core.callbacks import CallbackManager
from llama_index.core.llms.llm import LLM
from llama_index.core.memory import ChatMemoryBuffer
from llama_index.core.memory.chat_memory_buffer import ChatMemoryBuffer
from llama_index.core.memory.types import BaseMemory
from llama_index.core.objects.base import ObjectRetriever
from llama_index.core.tools import BaseTool, FunctionTool
from llama_index.core.tools.types import ToolMetadata
from llama_index.llms.openai.utils import OpenAIToolCall

from qai.agent.tools.types import StringEnum


class AgentType(StringEnum):
    campaign = "campaign"
    database = "database"
    conversation = "conversation"
    website_navigator = "website_navigator"
    # new_workflow = "new_workflow"
    # activate_existing_workflow = "activate_existing_workflow"


class FindAgentModel(BaseModel):
    sentence: str = Field(description="The user sentence to find the correct agent for.")
    agent_type: AgentType = Field(description="Which agent the user action belongs to.")


def find_agent(sentence: str, agent_type: AgentType) -> AgentType:
    print(f"Found agent for sentence: {sentence} agent_type: {agent_type}")

    # return {"sentence": sentence, "agent_type": agent_type}
    return agent_type


class FindAgentWorker(OpenAIAgent):
    """An Agent to find the best agent given a user sentence."""

    name: str = "find_agent"
    description: str = (
        f"Find the correct agent given a user sentence. "
        f"Only use one of the following agent types: [{','.join([str(e) for e in AgentType])}]"
    )

    @classmethod
    def create(
        cls,
        nametools: Sequence[BaseTool] = None,
        tools: Optional[List[BaseTool]] = None,
        tool_retriever: Optional[ObjectRetriever[BaseTool]] = None,
        llm: Optional[LLM] = None,
        chat_history: Optional[List[ChatMessage]] = None,
        memory: Optional[BaseMemory] = None,
        memory_cls: Type[BaseMemory] = ChatMemoryBuffer,
        verbose: bool = True,
        max_function_calls: int = DEFAULT_MAX_FUNCTION_CALLS,
        default_tool_choice: str = "auto",
        callback_manager: Optional[CallbackManager] = None,
        system_prompt: Optional[str] = None,
        prefix_messages: Optional[List[ChatMessage]] = None,
        tool_call_parser: Optional[Callable[[OpenAIToolCall], Dict]] = None,
        *args,
        **kwargs: Any,
    ):
        find_agent_tool = FunctionTool.from_defaults(
            name=cls.name,
            fn=find_agent,
            tool_metadata=ToolMetadata(
                name=cls.name,
                description=cls.description,
                fn_schema=FindAgentModel,
            ),
        )
        agent = FindAgentWorker.from_tools(
            llm=llm,
            tools=[find_agent_tool],
            max_function_calls=max_function_calls,
            default_tool_choice=default_tool_choice,
            callback_manager=callback_manager,
            prefix_messages=prefix_messages,
            tool_call_parser=tool_call_parser,
            chat_history=chat_history,
            memory=memory,
            memory_cls=memory_cls,
            verbose=verbose,
        )

        return agent
